==> itc-foundation-0.1.0.tgz: package/src/lang.js <==
const NotImplemented = new Error('not implemented');

/**
 * Works like PHP's empty() function
 * @param mixed value
 * @return {Boolean}
 */
const empty = (value) => {
  if (!value) {
    return value === undefined || !isNaN(value);
  }

  if (typeof value === 'number') {
    return value === 0;
  }

  if (typeof value.length === 'number') {
    return !value.length;
  }

  if (typeof value.size === 'number') {
    return !value.size;
  }

  if (typeof value === 'object') {
    for (let k in value) {
      if (value.hasOwnProperty(k)) {
        return false;
      }
    }

    return true;
  }

  return false;
};

/**
 * Prototype chaining function
 * @param {Function} subclass
 * @param {Function} superclass
 * @param {Object} augments
 * @return {Function} the subclass
 */
const extend = (subclass, superclass, augments) => {
  let Surrogate = function() {};

  Surrogate.prototype = superclass.prototype;
  subclass.prototype = new Surrogate();
  subclass.prototype.constructor = subclass;

  assign(subclass.prototype, augments || {}); 
       
  return subclass;
};

/**
 * Works like Object.assign()
 */
const assign = (...donors) => {
  let recipient = donors.shift();

  let augment = (recipient, donor) => {
    for (let k in donor) {
      if (donor.hasOwnProperty(k)) {
        recipient[k] = donor[k];
      }
    }
    return recipient;
  };

  return donors.reduce(augment, recipient);
};

/**
 * Return a concrete Iterator instance
 * @param {Array|Object|Map|Set} collection
 * @return {Iterator}
 */
const iterator = (collection) => {
  if (collection instanceof Iterator) {
    return collection;
  }
  if (Array.isArray(collection)) {
    return new ArrayIterator(collection);
  }
  if (collection instanceof Map) {
    return new MapIterator(collection);
  }
  if (collection instanceof Set) {
    return new SetIterator(collection);
  }
  if (collection && typeof collection === 'object') {
    return new ObjectIterator(collection);
  }

  throw new Error('expected iterable argument');
}

/**
 * Return an array of keys in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const keys = (collection) => iterator(collection).keys();

/**
 * Return an array of values in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const values = (collection) => iterator(collection).values();

/**
 * Return the size of the collection
 * @param {Array|Object|Map|Set} collection
 * @return Number
 */
const size = (collection) => iterator(collection).size();

/**
 * Return a representation of the collection as a list of two-element tuples
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const entries = (collection) => iterator(collection).entries();

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial
 * @return mixed
 */
const reduce = (...args) => reduce.until(...args);

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial - the initial carry
 * @param {Function} until - reduce until this function returns true
 * @return mixed
 */
reduce.until = (collection, fn, initial, until=()=>false) => {
  let pairs = entries(collection);
  let out = initial;

  while (pairs.length) {
    let [k, v] = pairs.shift();

    if (until(out, v, k, collection)) {
      break;
    }

    out = fn(out, v, k, collection);
  }

  return out;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - map function
 * @return {Array}
 */
const map = (...args) => map.until(...args);

/**
 * @param {Array}Object|Map|Set} collection
 * @param {Function} fn - map function
 * @param {Function} until - reduce until this function returns true
 * @return {Array}
 */
map.until = (collection, fn, until=()=>false) => {
  let reducer = (out, ...args) => {
    out.push(fn(...args));
    return out;
  };

  let test = (out, ...args) => until(...args);
  return reduce.until(collection, reducer, [], test);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @return void
 */
const each = (...args) => each.until(...args);

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @param {Function} until - iterate until this function returns true
 * @return void
 */
each.until = (collection, fn, until=()=>false) => {
  map.until(collection, fn, until);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - filter function
 * @return {Array}
 */
const filter = (collection, fn) => {
  return entries(collection).filter((entry) => {
    let [k, v] = entry;
    return fn(v, k, collection);
  });
};

/**
 * Return the first value in the collection that passes a truth test
 * Return the very first value in the collection if a truth test is not supplied
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - optional truth test
 * @return mixed
 */
const first = (collection, fn=v=>true) => {
  for (let [k, v] in entries(collection)) {
    if (fn(v, k, collection)) {
      return v;
    }
  }
  return null;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @return {Array|Object|Map|Set)
 */
const reverse = (collection) => iterator(collection).reverse();

export default {
  empty,
  extend,
  assign,
  iterator,
  keys,
  values,
  size,
  entries,
  reduce,
  each,
  map,
  filter,
  first,
  reverse,
}

////////////////////////////////////////////////////////////////////////////////


class Iterator {

  /**
   * Return an Array
   * @param void
   * @return {Array}
   */
  keys () {
    throw NotImplemented;
  }

  /**
   * Return an Array
   * @param void
   * @return {Array}
   */
  values () {
    throw NotImplemented;
  }

  /**
   * Return an Array of two-element tuples: [key, value]
   * @param void
   * @return {Array}
   */
  entries () {
    throw NotImplemented;
  }

  /**
   * @param void
   * @return {Number}
   */
  size () {
    throw NotImplemented;
  }

  /**
   * Return a reversed representation of the collection
   * @param void
   * @return mixed
   */
  reverse () {
    throw NotImplemented;
  }
}

class ArrayIterator extends Iterator {

  /**
   * @param {Array} array
   */
  constructor (array) {
    super();
    this.array = array;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push(i);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  values () {
    let {array} = this;
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push(array[i]);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let {array} = this;
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push([i, array[i]]);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.array.length;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    return this.array.reverse();
  }
}

class ObjectIterator extends Iterator {

  /**
   * @param {Object} object
   */
  constructor (object) {
    super();
    this.object = object;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push(k);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  values () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push(object[k]);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push([k, object[k]]);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.keys().length;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    return this.entries().reverse().reduce((out, v, k) => out[k] = v, {});
  }
}

class ES6CollectionIterator extends Iterator {

  /**
   * @param {Map|Set} collection
   */
  constructor (collection) {
    super();
    this.collection = collection;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let keys = [];
    this.collection.forEach((v, k) => keys.push(k));
    return keys;
  }

  /**
   * @inheritdoc
   */
  values () {
    let values = [];
    this.collection.forEach((v, k) => values.push(v));
    return values;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let entries = [];
    this.collection.forEach((v, k) => entries.push([k, v]));
    return entries;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.collection.size;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    let constructor = this.collection.constructor;
    return new constructor(this.entries().reverse());
  }
}

class MapIterator extends ES6CollectionIterator {}
class SetIterator extends ES6CollectionIterator {}

==> itc-foundation-0.2.1.tgz: package/src/lang.js <==
const NotImplemented = new Error('not implemented');

/**
 * Works like PHP's empty() function
 * @param mixed value
 * @return {Boolean}
 */
const empty = (value) => {
  if (!value) {
    return value === undefined || !isNaN(value);
  }

  if (typeof value === 'number') {
    return value === 0;
  }

  if (typeof value.length === 'number') {
    return !value.length;
  }

  if (typeof value.size === 'number') {
    return !value.size;
  }

  if (typeof value === 'object') {
    for (let k in value) {
      if (value.hasOwnProperty(k)) {
        return false;
      }
    }

    return true;
  }

  return false;
};

/**
 * Prototype chaining function
 * @param {Function} subclass
 * @param {Function} superclass
 * @param {Object} augments
 * @return {Function} the subclass
 */
const extend = (subclass, superclass, augments) => {
  let Surrogate = function() {};

  Surrogate.prototype = superclass.prototype;
  subclass.prototype = new Surrogate();
  subclass.prototype.constructor = subclass;

  assign(subclass.prototype, augments || {}); 
       
  return subclass;
};

/**
 * Works like Object.assign()
 */
const assign = (...donors) => {
  let recipient = donors.shift();

  let augment = (recipient, donor) => {
    for (let k in donor) {
      if (donor.hasOwnProperty(k)) {
        recipient[k] = donor[k];
      }
    }
    return recipient;
  };

  return donors.reduce(augment, recipient);
};

/**
 * Return a concrete Iterator instance
 * @param {Array|Object|Map|Set} collection
 * @return {Iterator}
 */
const iterator = (collection) => {
  if (collection instanceof Iterator) {
    return collection;
  }
  if (Array.isArray(collection)) {
    return new ArrayIterator(collection);
  }
  if (collection instanceof Map) {
    return new MapIterator(collection);
  }
  if (collection instanceof Set) {
    return new SetIterator(collection);
  }
  if (collection && typeof collection === 'object') {
    return new ObjectIterator(collection);
  }

  throw new Error('expected iterable argument');
}

/**
 * Return an array of keys in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const keys = (collection) => iterator(collection).keys();

/**
 * Return an array of values in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const values = (collection) => iterator(collection).values();

/**
 * Return the size of the collection
 * @param {Array|Object|Map|Set} collection
 * @return Number
 */
const size = (collection) => iterator(collection).size();

/**
 * Return a representation of the collection as a list of key-value tuples:
 *
 *   [['key1', 'value1'], ['key2', 'value2'], ..., ['keyN', 'valueN']]
 *
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
const entries = (collection) => iterator(collection).entries();

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial
 * @return mixed
 */
const reduce = (...args) => reduce.until(...args);

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial - the initial carry
 * @param {Function} breakpoint - reduce until this function returns true
 * @return mixed
 */
reduce.until = (collection, fn, initial, breakpoint=()=>false) => {
  let pairs = entries(collection);
  let out = initial;

  while (pairs.length) {
    let [k, v] = pairs.shift();

    if (breakpoint(out, v, k, collection)) {
      break;
    }

    out = fn(out, v, k, collection);
  }

  return out;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - map function
 * @return {Array}
 */
const map = (...args) => map.until(...args);

/**
 * @param {Array}Object|Map|Set} collection
 * @param {Function} fn - map function
 * @param {Function} breakpoint - reduce until this function returns true
 * @return {Array}
 */
map.until = (collection, fn, breakpoint=()=>false) => {
  let reducer = (out, ...args) => {
    out.push(fn(...args));
    return out;
  };

  let test = (out, ...args) => breakpoint(...args);
  return reduce.until(collection, reducer, [], test);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @return void
 */
const each = (...args) => each.until(...args);

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @param {Function} breakpoint - iterate until this function returns true
 * @return void
 */
each.until = (collection, fn, breakpoint=()=>false) => {
  map.until(collection, fn, breakpoint);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - filter function
 * @return {Array}
 */
const filter = (collection, fn) => {
  let pairs = entries(collection);
  let out = [];

  while (pairs.length) {
    let [k, v] = pairs.shift();
    fn(v, k, collection) && out.push(v);
  }

  return out;
};

/**
 * Return the first value in the collection that passes a truth test
 * Return the very first value in the collection if a truth test is not supplied
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - optional truth test
 * @return mixed
 */
const first = (collection, fn=v=>true) => {
  let pairs = entries(collection);

  while (pairs.length) {
    let [k, v] = pairs.shift();

    if (fn(v, k, collection)) {
      return v;
    }
  }

  return null;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @return {Array|Object|Map|Set)
 */
const reverse = (collection) => iterator(collection).reverse();

export default {
  empty,
  extend,
  assign,
  keys,
  values,
  size,
  entries,
  reduce,
  each,
  map,
  filter,
  first,
  reverse,
}

////////////////////////////////////////////////////////////////////////////////

class Iterator {

  /**
   * Return an Array
   * @param void
   * @return {Array}
   */
  keys () {
    throw NotImplemented;
  }

  /**
   * Return an Array
   * @param void
   * @return {Array}
   */
  values () {
    throw NotImplemented;
  }

  /**
   * Return an Array of two-element tuples: [key, value]
   * @param void
   * @return {Array}
   */
  entries () {
    throw NotImplemented;
  }

  /**
   * @param void
   * @return {Number}
   */
  size () {
    throw NotImplemented;
  }

  /**
   * Return a reversed representation of the collection
   * @param void
   * @return mixed
   */
  reverse () {
    throw NotImplemented;
  }
}

class ArrayIterator extends Iterator {

  /**
   * @param {Array} array
   */
  constructor (array) {
    super();
    this.array = array;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push(i);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  values () {
    let {array} = this;
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push(array[i]);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let {array} = this;
    let out = [];
    let size = this.size();
    let i = 0;

    while (i < size) {
      out.push([i, array[i]]);
      i += 1;
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.array.length;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    return this.array.reverse();
  }
}

class ObjectIterator extends Iterator {

  /**
   * @param {Object} object
   */
  constructor (object) {
    super();
    this.object = object;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push(k);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  values () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push(object[k]);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let {object} = this;
    let out = [];

    for (let k in object) {
      object.hasOwnProperty(k) && out.push([k, object[k]]);
    }

    return out;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.keys().length;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    return this.entries().reverse().reduce((out, v, k) => out[k] = v, {});
  }
}

class ES6CollectionIterator extends Iterator {

  /**
   * @param {Map|Set} collection
   */
  constructor (collection) {
    super();
    this.collection = collection;
  }

  /**
   * @inheritdoc
   */
  keys () {
    let keys = [];
    this.collection.forEach((v, k) => keys.push(k));
    return keys;
  }

  /**
   * @inheritdoc
   */
  values () {
    let values = [];
    this.collection.forEach((v, k) => values.push(v));
    return values;
  }

  /**
   * @inheritdoc
   */
  entries () {
    let entries = [];
    this.collection.forEach((v, k) => entries.push([k, v]));
    return entries;
  }

  /**
   * @inheritdoc
   */
  size () {
    return this.collection.size;
  }

  /**
   * @inheritdoc
   */
  reverse () {
    let constructor = this.collection.constructor;
    return new constructor(this.entries().reverse());
  }
}

class MapIterator extends ES6CollectionIterator {}
class SetIterator extends ES6CollectionIterator {}
